import copy
import math
import operator
import pickle
import sys
from typing import Any, Generator


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton,
                                        cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class _SELF(metaclass=Singleton):
    __slots__ = ()

    def __repr__(self):
        return "self"


class _NOTSET(metaclass=Singleton):
    __slots__ = ()

    def __repr__(self):
        return 'N/A'


class _UNKNOW(metaclass=Singleton):
    __slots__ = ()

    def __repr__(self) -> str:
        return "Unknow"


class _DELETE(metaclass=Singleton):
    __slots__ = ()

    def __repr__(self):
        return 'Delete'


SELF = _SELF()
NOTSET = _NOTSET()
UNKNOW = _UNKNOW()
DELETE = _DELETE()


def flattenDictIter(d: dict,
                    prefix: list = []
                    ) -> Generator[tuple[str, Any], None, None]:
    for k in d:
        if isinstance(d[k], dict):
            yield from flattenDictIter(d[k], prefix=[*prefix, k])
        else:
            yield '.'.join(prefix + [k]), d[k]


def flattenDict(d: dict[str, Any]) -> dict[str, Any]:
    return {k: v for k, v in flattenDictIter(d)}


def foldDict(d: dict[str, Any]) -> dict[str, Any]:
    ret = {}

    for k, v in d.items():
        keys = k.split('.')
        d = ret
        parent = None

        for key in keys[:-1]:
            if not isinstance(d, dict):
                parent[0][parent[1]] = {SELF: d}
                d = parent[0][parent[1]]
            if key not in d:
                d[key] = dict()
            parent = d, key
            d = d[key]
        if not isinstance(d, dict):
            parent[0][parent[1]] = {SELF: d}
            d = parent[0][parent[1]]
        if keys[-1] in d and isinstance(d[keys[-1]], dict):
            d[keys[-1]][SELF] = v
        else:
            d[keys[-1]] = v
    return ret


class Update():
    __slots__ = ('o', 'n')

    def __init__(self, o, n):
        self.o = o
        self.n = n

    def __repr__(self):
        return f"Update: {self.o!r} ==> {self.n!r}"

    def __eq__(self, o: object) -> bool:
        return isinstance(o, Update) and self.n == o.n


class Create():
    __slots__ = ('n', 'replace')

    def __init__(self, n, replace=False):
        """
        Create a new node

        :param n: the new node
        :param replace: if True, replace the node if it already exists
        """
        self.n = n
        self.replace = replace

    def __repr__(self):
        if self.replace:
            return f"Replace: {self.n!r}"
        else:
            return f"Create: {self.n!r}"

    def __eq__(self, o: object) -> bool:
        return isinstance(
            o, Create) and self.n == o.n and self.replace == o.replace


def _eq(a, b):
    import numpy as np

    if isinstance(a, np.ndarray):
        return np.array_equal(a, np.asarray(b))

    try:
        return a == b
    except:
        pass
    if isinstance(a, (list, tuple)):
        return len(a) == len(b) and all(_eq(a[i], b[i]) for i in range(len(a)))
    if isinstance(a, dict):
        return set(a.keys()) == set(b.keys()) and all(
            _eq(a[k], b[k]) for k in a)

    try:
        return pickle.dumps(a) == pickle.dumps(b)
    except:
        return False


def diff(d1: dict, d2: dict) -> dict:
    """
    Compute the difference between two dictionaries

    Args:
        d1: the original dictionary
        d2: the new dictionary

    Returns:
        a dictionary containing the difference between d1 and d2
    """
    ret = {}
    for k in d2:
        if k in d1:
            if isinstance(d2[k], type(d1[k])) and _eq(d1[k], d2[k]):
                pass
            elif isinstance(d1[k], dict) and isinstance(d2[k], dict):
                ret[k] = diff(d1[k], d2[k])
            else:
                ret[k] = Update(d1[k], d2[k])
        else:
            ret[k] = Create(d2[k])
    for k in d1:
        if k not in d2:
            ret[k] = DELETE
    return ret


def patch(source, diff, in_place=False):
    """
    Patch a dictionary with a diff

    Args:
        source: the original dictionary
        diff: the diff
        in_place: if True, patch the source dictionary in place
    
    Returns:
        the patched dictionary
    """
    if in_place:
        ret = source
    else:
        ret = copy.copy(source)
    for k, v in diff.items():
        if isinstance(v, dict):
            ret[k] = patch(source[k], v, in_place=in_place)
        else:
            if isinstance(v, Update):
                ret[k] = v.n
            elif isinstance(v, Create):
                if v.replace or k not in ret:
                    ret[k] = v.n
                else:
                    update_tree(ret[k], v.n)
            elif v is DELETE:
                del ret[k]
            else:
                raise ValueError(f"Unsupported patch: {v!r}")
    return ret


def merge(diff1, diff2, origin=None):
    """
    Merge two diffs

    Args:
        diff1: the first diff
        diff2: the second diff
        origin: the original dictionary

    Returns:
        the merged diff
    """
    if origin is not None:
        updated = patch(patch(origin, diff1), diff2)
        return diff(origin, updated)

    ret = {}
    for k, v in diff1.items():
        if k in diff2:
            v2 = diff2[k]
            if isinstance(v, dict) and isinstance(v2, dict):
                d = merge(v, v2)
                if d:
                    ret[k] = d
            else:
                if isinstance(v, Update) and isinstance(v2, Update):
                    ret[k] = Update(v.o, v2.n)
                elif isinstance(v, Create) and isinstance(v2, dict):
                    ret[k] = Create(patch(copy.copy(v.n), v2, True))
                elif isinstance(v, Create) and isinstance(v2, Update):
                    ret[k] = Create(v2.n)
                elif isinstance(v, Create) and v2 is DELETE:
                    pass
                elif v2 is DELETE:
                    ret[k] = DELETE
                elif v is DELETE and isinstance(v2, Create):
                    if isinstance(v2.n, dict):
                        ret[k] = Create(v2.n, replace=True)
                    else:
                        ret[k] = Update(UNKNOW, v2.n)
                elif isinstance(v2, Create) and v2.replace:
                    ret[k] = v2
                else:
                    raise ValueError(f"Unsupported merge: {v!r} {v2!r}")
        else:
            ret[k] = v
    for k, v in diff2.items():
        if k not in diff1:
            ret[k] = v
    return ret


def print_diff(d, limit=None, offset=0, file=sys.stdout, ignores=None):
    """
    Print a diff

    Args:
        d: the diff
        limit: the maximum number of lines to print
        offset: the offset of the first line
        file: the file to print to
        ignores: ignore keys starting with this prefix
    """
    count = 0
    for i, (k, v) in enumerate(flattenDictIter(d)):
        if count >= offset:
            if ignores is not None and k.startswith(ignores):
                continue
            print(f"{k:40}", v, file=file)
            count += 1
            if limit is not None and count >= limit:
                break


def update_tree(result, updates):
    for k, v in updates.items():
        if isinstance(v, dict):
            if k not in result or not isinstance(result[k], dict):
                result[k] = {}
            update_tree(result[k], v)
        else:
            result[k] = v
    return result


def queryref_tree(q, keys, dct, prefix=[], chain=None):
    q = q[1:]
    if q.startswith('.'):
        while q.startswith('.'):
            keys.pop()
            q = q[1:]
        q = '.'.join(keys + [q])

    if chain is None:
        chain = [q]
    elif q in chain:
        raise KeyError(f'Circular reference: {chain+[q]}')

    return query_tree(q, dct, prefix=prefix, eval=True, chain=chain + [q])


def query_tree(q, dct, prefix=[], eval=True, chain=None):
    ret = dct
    keys = q.split('.')
    for i, key in enumerate(keys):
        if key not in ret:
            return (NOTSET, '.'.join(prefix + keys[:i + 1]))
        ret = ret[key]
    if isinstance(ret, str) and eval:
        if ret.startswith('$'):
            return queryref_tree(ret, keys, dct, prefix=prefix, chain=chain)
        elif ret.startswith('&'):
            return eval_expr(ret[1:], Env(dct,
                                          keys,
                                          prefix=prefix,
                                          chain=chain))
    return ret


class Env():

    def __init__(self, dct=None, keys=None, prefix=[], chain=None):
        self.dct = dct if dct is not None else {}
        self.keys = keys if keys is not None else []
        self.chain = chain
        self.prefix = prefix

    def get(self, name):
        return queryref_tree(name[1:],
                             self.keys,
                             self.dct,
                             prefix=self.prefix,
                             chain=self.chain)


internal_functions = {
    '**': operator.pow,
    '+': operator.add,
    '-': operator.sub,
    '*': operator.mul,
    '/': operator.truediv,
    '//': operator.floordiv,
    '%': operator.mod,
    '&': operator.and_,
    '|': operator.or_,
    '^': operator.xor,
    '~': operator.invert,
    '<<': operator.lshift,
    '>>': operator.rshift,
    'getitem': operator.getitem,
    'contains': operator.contains,
    'not': operator.not_,
    'getattr': getattr,
    'abs': abs,
    'min': min,
    'max': max,
    'sum': sum,
    'len': len,
    'any': any,
    'all': all,
    'ord': ord,
    'chr': chr,
    'bin': bin,
    'oct': oct,
    'hex': hex,
    'int': int,
    'round': round,
    'real': lambda x: x.real,
    'imag': lambda x: x.imag,
    'sqrt': math.sqrt,
    'sin': math.sin,
    'cos': math.cos,
    'tan': math.tan,
    'asin': math.asin,
    'acos': math.acos,
    'atan': math.atan,
    'atan2': math.atan2,
    'sinh': math.sinh,
    'cosh': math.cosh,
    'tanh': math.tanh,
    'asinh': math.asinh,
    'acosh': math.acosh,
    'atanh': math.atanh,
    'ceil': math.ceil,
    'floor': math.floor,
    'trunc': math.trunc,
    'log10': math.log10,
    'log': math.log,
    'exp': math.exp,
}


def eval_expr(expression, env=None, functions=None):
    from pyparsing import (Forward, Literal, ParseResults, Suppress, Word,
                           alphanums, delimitedList, infixNotation, oneOf,
                           opAssoc, pyparsing_common)
    if functions is None:
        functions = internal_functions

    def lookup(name):
        return env.get(name)

    def apply(fun, *args):
        return functions[fun](*args)

    def eval_expr(t):
        if not isinstance(t, ParseResults):
            return t
        if t[0] == '-' and len(t) == 2:
            return -t[1]
        return apply(t[1], t[0], t[2])

    # Define pyparsing grammar
    expr = Forward()

    complex_ = (pyparsing_common.number +
                'j').setParseAction(lambda s, l, t: complex(t[0]))
    number = pyparsing_common.number | complex_

    const = oneOf('pi e').setParseAction(lambda s, l, t: {
        'pi': math.pi,
        'e': math.e
    }.get(t[0]))

    variable = Word("$", alphanums +
                    "._").setParseAction(lambda s, l, t: lookup(t[0]))

    bracket = Suppress('(') + expr + Suppress(')')
    bracket.setParseAction(lambda s, l, t: t[0])

    operand = const | variable | number | bracket

    func = pyparsing_common.identifier
    func_call = func + Suppress("(") + delimitedList(expr) + Suppress(")")
    func_call.setParseAction(lambda s, l, t: apply(t[0], *t[1:]))

    term = operand | func_call

    expr << infixNotation(term, [
        (Literal('**'), 2, opAssoc.RIGHT),
        (Literal('~'), 1, opAssoc.RIGHT),
        (Literal('-'), 1, opAssoc.RIGHT),
        (oneOf('* / // %'), 2, opAssoc.LEFT),
        (oneOf('+ -'), 2, opAssoc.LEFT),
        (oneOf('>> <<'), 2, opAssoc.LEFT),
        (Literal('&'), 2, opAssoc.LEFT),
        (Literal('^'), 2, opAssoc.LEFT),
        (Literal('|'), 2, opAssoc.LEFT),
    ])

    expr.setParseAction(lambda s, l, t: eval_expr(t[0]))

    parsed = expr.parseString(expression, parseAll=True)
    return parsed[0]


def sorted_tree(dct, *, keys=None):
    if keys is None or callable(keys):
        key = keys
    elif isinstance(keys, list):
        key = keys[0]
        if len(keys) > 1:
            keys = keys[1:]
        else:
            keys = None
    elif (isinstance(keys, tuple) and len(keys) == 2
          and isinstance(keys[1], dict)):
        key = keys[0]
        keys = keys[1]
    else:
        raise Exception(f"Unsupported keys: {keys!r}")

    if isinstance(dct, dict):
        if isinstance(keys, dict):
            default = keys.get('default', None)
            return {
                k: sorted_tree(dct[k], keys=keys.get(k, default))
                for k in sorted(dct.keys(), key=key)
            }
        else:
            return {
                k: sorted_tree(dct[k], keys=keys)
                for k in sorted(dct.keys(), key=key)
            }
    elif isinstance(dct, set):
        if isinstance(keys, dict):
            default = keys.get('default', None)
            return set([
                sorted_tree(v, keys=keys.get(v, default))
                for v in sorted(list(dct), key=key)
            ])
        else:
            return set([
                sorted_tree(v, keys=keys) for v in sorted(list(dct), key=key)
            ])
    else:
        return dct



import numpy as np
from scipy.linalg import expm


def _rot(theta, v):
    x, y, z = np.asarray(v) / np.sqrt(np.sum(np.asarray(v)**2))
    c, s = np.cos(theta), np.sin(theta)

    return np.array([[
        c + (1 - c) * x**2, (1 - c) * x * y - s * z, (1 - c) * x * z + s * y
    ], [(1 - c) * x * y + s * z, c + (1 - c) * y**2, (1 - c) * y * z - s * x],
                     [(1 - c) * x * z - s * y, (1 - c) * y * z + s * x,
                      c + (1 - c) * z**2]])


def rot_round(x, y, z, v=[0, 0, 1], theta=0, c=[0, 0, 0]):
    ret = (np.array([x, y, z]).T - np.array(c)) @ _rot(theta, v) + np.array(c)
    return ret.T


def projection(x, y, z, y0=1):
    d = np.sqrt(x**2 + y**2 + z**2)
    return x * y0 / y, z * y0 / y, d



from .storage import Report, save_item

__current_config_id = None


def _query_config(name: str, default=None):
    import pickle

    try:
        with open('parameters.pkl', 'rb') as f:
            parameters = pickle.load(f)
    except:
        parameters = {}

    return parameters.get(name, default)


def _update_config(updates):
    import pickle

    try:
        with open('parameters.pkl', 'rb') as f:
            parameters = pickle.load(f)
    except:
        parameters = {}

    for k, v in updates.items():
        parameters[k] = v

    with open('parameters.pkl', 'wb') as f:
        pickle.dump(parameters, f)


def _export_config() -> dict:
    import pickle

    try:
        with open('parameters.pkl', 'rb') as f:
            parameters = pickle.load(f)
    except:
        parameters = {}

    return parameters


def obey_the_oracle(report: Report, data_path):
    global __current_config_id
    update_config(report.oracle)
    cfg = export_config()
    __current_config_id = save_item(cfg, data_path)


def update_parameters(report: Report, data_path):
    global __current_config_id
    update_config(report.parameters)
    cfg = export_config()
    __current_config_id = save_item(cfg, data_path)


def current_config(data_path):
    global __current_config_id
    if __current_config_id is None:
        cfg = export_config()
        __current_config_id = save_item(cfg, data_path)
    return __current_config_id


query_config = _query_config
update_config = _update_config
export_config = _export_config


def set_config_api(query_method, update_method, export_method):
    """
    Set the query and update methods for the config.

    Args:
        query_method: The query method.
            the method should take a key and return the value.
        update_method: The update method.
            the method should take a dict of updates.
        export_method: The export method.
            the method should return a dict of the config.
    """
    global query_config, update_config, export_config

    query_config = query_method
    update_config = update_method
    export_config = export_method

    return query_config, update_config, export_config



import asyncio
import random

import zmq
import zmq.asyncio


async def handle_client(socket, identity, message):
    print(f"Received request from {identity}: {message.decode()}")
    # 随机延时 0 到 3 秒
    await asyncio.sleep(random.uniform(0, 3))
    task_id = random.randint(1000, 9999)  # 随机生成一个任务 ID
    await socket.send_multipart([identity, f"Task ID: {task_id}".encode()])
    print(f"Sent Task ID {task_id} to {identity}")


async def server():
    context = zmq.asyncio.Context()
    socket = context.socket(zmq.ROUTER)
    socket.bind("tcp://*:5555")

    while True:
        try:
            identity, message = await socket.recv_multipart()
            asyncio.create_task(handle_client(socket, identity, message))
        except Exception as e:
            print(f"An error occurred: {e}")
            break

    socket.close()
    context.term()


if __name__ == "__main__":
    asyncio.run(server())



import pathlib
import pickle

import click
import dill
import matplotlib.pyplot as plt
import numpy as np

from .qdat import draw as draw_qdat

default_draw_methods = {
    '.qdat': draw_qdat,
}


def load_data(fname):
    try:
        from home.hkxu.tools import get_record_by_id
        record_id = int(str(fname))
        return get_record_by_id(record_id).data
    except:
        pass
    with open(fname, 'rb') as f:
        try:
            data = pickle.load(f)
        except:
            f.seek(0)
            data = dill.load(f)
    return data


def draw_common(data):
    try:
        script = data['meta']['plot_script']
        assert script.strip()
        global_namespace = {'plt': plt, 'np': np, 'result': data}
        exec(script, global_namespace)
    except:
        from home.hkxu.tools import plot_record
        plot_record(data['meta']['id'])


def draw_error(data, text="No validate plot script found"):
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.text(0.5, 0.5, text, ha='center', va='center')
    ax.set_axis_off()
    return fig


@click.command()
@click.argument('fname', default='')
def plot(fname):
    """Plot the data in the file."""
    try:
        fname = pathlib.Path(fname)
        data = load_data(fname)
        try:
            draw_common(data)
        except:
            default_draw_methods.get(fname.suffix, draw_error)(data)
    except FileNotFoundError:
        draw_error(None, text=f"File {fname} not found.")
    except pickle.UnpicklingError:
        draw_error(None, text=f"File {fname} is not a pickle file.")

    plt.show()


if __name__ == '__main__':
    plot()



from matplotlib.backends.qt_compat import QT_API, QtCore, QtWidgets

if QT_API in ['PySide6', 'PyQt6']:
    AlignRight = QtCore.Qt.AlignmentFlag.AlignRight
    BottomDockWidgetArea = QtCore.Qt.DockWidgetArea.BottomDockWidgetArea
    ScrollBarAlwaysOn = QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOn
    ScrollBarAlwaysOff = QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff
    TopDockWidgetArea = QtCore.Qt.DockWidgetArea.TopDockWidgetArea
elif QT_API in ['PyQt5', 'PySide2']:
    AlignRight = QtCore.Qt.AlignRight
    BottomDockWidgetArea = QtCore.Qt.BottomDockWidgetArea
    ScrollBarAlwaysOn = QtCore.Qt.ScrollBarAlwaysOn
    ScrollBarAlwaysOff = QtCore.Qt.ScrollBarAlwaysOff
    TopDockWidgetArea = QtCore.Qt.TopDockWidgetArea
else:
    raise AssertionError(f"Unexpected QT_API: {QT_API}")



import dill
from sqlalchemy import Column, ForeignKey, Integer, String
from sqlalchemy.orm import relationship

from .base import Base
from .file import File
from .tag import Tag, has_tags


@has_tags
class Config(Base):
    __tablename__ = 'configs'

    id = Column(Integer, primary_key=True)
    name = Column(String)
    file_id = Column(Integer, ForeignKey('files.id'))
    meta_id = Column(Integer, ForeignKey('files.id'))

    file = relationship("File", foreign_keys=[file_id])

    @property
    def data(self) -> dict:
        result = dill.loads(self.file.read())
        return result

    @data.setter
    def data(self, data: dict):
        self.file.write(dill.dumps(data))



import subprocess
import sys


def run_detached_with_terminal(executable_path):
    """
    启动可执行文件并在新终端窗口中保持运行，Python退出后进程仍存在。
    适用于Windows、Linux和macOS。
    """
    try:
        if sys.platform == 'win32':
            # Windows：使用start命令启动新cmd窗口
            cmd = f'start cmd /k "{executable_path}"'
            subprocess.Popen(cmd, shell=True)
        elif sys.platform == 'darwin':
            # macOS：通过AppleScript在Terminal中执行命令
            escaped_path = executable_path.replace('"', r'\"')
            script = f'tell application "Terminal" to do script "{escaped_path}"'
            subprocess.Popen(['osascript', '-e', script],
                             start_new_session=True)
        else:
            # Linux：尝试gnome-terminal或xterm
            try:
                subprocess.Popen(['gnome-terminal', '--', executable_path],
                                 start_new_session=True)
            except FileNotFoundError:
                subprocess.Popen(['xterm', '-e', executable_path],
                                 start_new_session=True)
    except Exception as e:
        print(f"启动失败: {e}")
        sys.exit(1)



import asyncio
from collections import deque
from datetime import timedelta
from time import monotonic

from blinker import Signal

try:
    import ipywidgets as widgets
    from IPython.display import display
except:
    pass


class Progress:
    sma_window = 10  # Simple Moving Average window

    def __init__(self, *, max=10):
        self.start_ts = monotonic()
        self._ts = self.start_ts
        self._xput = deque(maxlen=self.sma_window)
        self._max = max
        self.index = 0
        self.updated = Signal()
        self.finished = Signal()
        self.status = "running"

    @property
    def progress(self):
        return min(1.0, self.index / self.max)

    @property
    def remaining(self):
        return max(self.max - self.index, 0)

    @property
    def percent(self):
        return self.progress * 100

    @property
    def elapsed(self):
        if self.status != "running":
            return self._ts - self.start_ts
        return monotonic() - self.start_ts

    @property
    def eta(self):
        if self.status != "running":
            return 0
        avg = sum(self._xput) / len(self._xput)
        dt = monotonic() - self._ts
        if dt > avg:
            avg = 0.9 * avg + 0.1 * dt
            return avg * self.remaining
        else:
            return avg * self.remaining - dt

    @property
    def max(self):
        return self._max

    @max.setter
    def max(self, x):
        self._max = x
        self.updated.send(self)

    def next(self, n=1):
        assert n > 0
        now = monotonic()
        dt = now - self._ts
        self._xput.append(dt / n)
        self._ts = now
        self.index = self.index + n
        self.updated.send(self)

    def goto(self, index):
        incr = index - self.index
        self.next(incr)

    def finish(self, success=True):
        self._ts = monotonic()
        if success:
            self.status = "finished"
        else:
            self.status = "failure"
        self.finished.send(self, success=success)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            self.finish(True)
        else:
            self.finish(False)

    def iter(self, it, max=None):
        if max is None:
            try:
                self.max = len(it)
            except TypeError:
                self.max = 0
        else:
            self.max = max

        with self:
            for x in it:
                yield x
                self.next()

    def __repr__(self):
        if self._ts == self.start_ts:
            eta_td = '--:--:--'
        else:
            eta_td = timedelta(seconds=round(self.eta))
        return (
            f"({self.index}/{self.max}) {self.percent:.0f}%"
            f" Used time: {timedelta(seconds=round(self.elapsed))} Remaining time: {eta_td}"
            f" {self.status}")


class ProgressBar():

    def listen(self, progress: Progress):
        self.progress = progress
        self.progress.updated.connect(self.update)
        self.progress.finished.connect(self.finish)

    def update(self, sender: Progress):
        raise NotImplementedError()

    def finish(self, sender: Progress, success: bool):
        self.progress.updated.disconnect(self.update)
        self.progress.finished.disconnect(self.finish)


class JupyterProgressBar(ProgressBar):

    def __init__(self, *, description='Progressing', hiden=False):
        self.description = description
        self.hiden = hiden

    def display(self):
        if self.hiden:
            return
        self.progress_ui = widgets.FloatProgress(value=0,
                                                 min=0,
                                                 max=100.0,
                                                 step=1,
                                                 description=self.description,
                                                 bar_style='')

        self.elapsed_ui = widgets.Label(value='Used time: 00:00:00')
        self.eta_ui = widgets.Label(value='Remaining time: --:--:--')
        self.ui = widgets.HBox(
            [self.progress_ui, self.elapsed_ui, self.eta_ui])
        display(self.ui)

    def listen(self, progress: Progress):
        super().listen(progress)
        self.update_regularly()

    def update_regularly(self, frequency=1):
        try:
            self.update(self.progress)
        except:
            pass
        asyncio.get_running_loop().call_later(frequency, self.update_regularly)

    def update(self, sender: Progress):
        if self.hiden:
            return
        self.progress_ui.value = sender.percent
        self.elapsed_ui.value = (
            f'({sender.index}/{sender.max}) '
            f'Used time: {timedelta(seconds=round(sender.elapsed))}')
        if sender.eta == sender.start_ts:
            self.eta_ui.value = f'Remaining time: --:--:--'
        else:
            self.eta_ui.value = f'Remaining time: {timedelta(seconds=round(sender.eta))}'

    def finish(self, sender: Progress, success: bool = True):
        if self.hiden:
            return
        if success:
            self.progress_ui.bar_style = 'success'
            self.progress_ui.value = 100.0
        else:
            self.progress_ui.bar_style = 'danger'
        super().finish(sender, success)



import ast
import base64
import hashlib
import lzma
import pickle
import re
import string
import textwrap
from typing import Any


def encode_mapping(mapping):
    mapping_bytes = lzma.compress(pickle.dumps(mapping))
    hash_str = hashlib.md5(mapping_bytes).hexdigest()[:8]
    mappping_code = '\n'.join(
        textwrap.wrap(base64.b64encode(mapping_bytes).decode(),
                      90,
                      initial_indent='    ',
                      subsequent_indent='    '))
    return hash_str, mappping_code


def decode_mapping(hash_str, mappping_code):
    mapping_bytes = base64.b64decode(mappping_code)
    if hash_str != hashlib.md5(mapping_bytes).hexdigest()[:8]:
        raise ValueError("Hash does not match")
    mapping = pickle.loads(lzma.decompress(mapping_bytes))
    return mapping


class TemplateTypeError(TypeError):
    pass


class TemplateKeyError(KeyError):
    pass


class TemplateVarExtractor(ast.NodeVisitor):

    def __init__(self, fname, mapping):
        self.var_func_def = (0, 0)
        self.variables = set()
        self.str_variables = set()
        self.replacements = {}
        self.fname = fname
        self.mapping = mapping

    def visit_Constant(self, node):
        if isinstance(node.value, str):
            self._process_string(node.value, node.lineno, node.col_offset,
                                 node.end_lineno, node.end_col_offset)

    def visit_JoinedStr(self, node):
        for value in node.values:
            if isinstance(value, ast.Constant) and isinstance(
                    value.value, str):
                self._process_string(value.value, value.lineno,
                                     value.col_offset, value.end_lineno,
                                     value.end_col_offset)
        self.generic_visit(node)

    def visit_FunctionDef(self, node):
        if node.name == 'VAR':
            self.var_func_def = (node.lineno, node.end_lineno)
        self.generic_visit(node)

    def visit_Call(self, node):
        if isinstance(node.func, ast.Name) and node.func.id == 'VAR':
            arg = node.args[0]
            if isinstance(arg, ast.Constant) and isinstance(arg.value, str):
                if arg.value not in self.mapping:
                    raise TemplateKeyError(
                        f"The variable '{arg.value}' is not provided in mapping. {self.fname}:{node.lineno}"
                    )
                self.variables.add(arg.value)
                # new_node = ast.Subscript(value=ast.Name(id="__VAR",
                #                                         ctx=ast.Load()),
                #                          slice=ast.Constant(value=arg.value),
                #                          ctx=ast.Load())
                # ast.fix_missing_locations(new_node)
                # new_source = ast.unparse(new_node)
                self.replacements[(node.lineno, node.end_lineno,
                                   node.col_offset,
                                   node.end_col_offset)] = ('VAR', arg.value,
                                                            None, None)
            else:
                raise SyntaxError(
                    f"Argument of VAR function must be a string. {self.fname}:{node.lineno}"
                )
        self.generic_visit(node)

    def _process_string(self, s: str, lineno: int, col_offset: int,
                        end_lineno: int, end_col_offset: int):
        """解析字符串内容，提取模板变量"""
        lines = s.split('\n')
        for offset, line in enumerate(lines):
            current_lineno = lineno + offset
            template = string.Template(line)
            for var_name in template.get_identifiers():
                if var_name not in self.mapping:
                    raise TemplateKeyError(
                        f"The variable '{var_name}' is not provided in mapping. {self.fname}:{current_lineno}"
                    )
                if not isinstance(self.mapping[var_name], str):
                    raise TemplateTypeError(
                        f"Mapping value for '{var_name}' must be a string. {self.fname}:{current_lineno}"
                    )
                self.str_variables.add(var_name)
                start, stop = 0, len(line)
                if current_lineno == lineno:
                    start = col_offset
                if current_lineno == end_lineno:
                    stop = end_col_offset
                self.replacements[(current_lineno, current_lineno, start,
                                   stop)] = ('STR', var_name, None, None)


def inject_mapping(source: str, mapping: dict[str, Any],
                   fname: str) -> list[tuple[str, int]]:
    hash_str, mapping_code = encode_mapping(mapping)

    tree = ast.parse(source)

    lines = source.splitlines()
    lines_offset = [0 for _ in range(len(lines))]

    extractor = TemplateVarExtractor(fname, mapping)
    extractor.visit(tree)

    # remove VAR function definition
    if extractor.var_func_def != (0, 0):
        for i in range(extractor.var_func_def[0] - 1,
                       extractor.var_func_def[1]):
            lines[i] = ''

    for (lineno, end_lineno, col_offset,
         end_col_offset), (kind, name, old_source,
                           new_source) in extractor.replacements.items():
        head = lines[lineno - 1][:col_offset + lines_offset[lineno - 1]]
        tail = lines[end_lineno - 1][end_col_offset +
                                     lines_offset[end_lineno - 1]:]
        length_of_last_line = len(lines[end_lineno - 1])
        content = lines[lineno - 1:end_lineno]
        content[0] = content[0].removeprefix(head)
        content[-1] = content[-1].removesuffix(tail)
        content = '\n'.join(content)

        if kind == 'STR':
            template = string.Template(content)
            formated_lines = template.substitute(mapping).splitlines()
            formated_lines[0] = head + formated_lines[0]
            formated_lines[-1] = formated_lines[-1] + tail
            if len(formated_lines) == 1:
                lines[lineno - 1] = formated_lines[0]
                lines_offset[lineno -
                             1] += len(lines[lineno - 1]) - length_of_last_line
            else:
                lines[lineno - 1:end_lineno] = formated_lines
                lines_offset[end_lineno - 1] += len(
                    lines[end_lineno - 1]) - length_of_last_line
        else:
            pattern = re.compile(r'VAR\s*\(\s*(["\'])(\w+)\1\s*\)')
            replacement = f'__VAR_{hash_str}' + r'[\1\2\1]'
            new_content = re.sub(pattern, replacement, content)

            if lineno == end_lineno:
                lines[lineno - 1] = head + new_content + tail
                lines_offset[lineno -
                             1] += len(lines[lineno - 1]) - length_of_last_line
            else:
                lines[lineno - 1] = head + new_content[:-1]
                for i in range(lineno, end_lineno - 1):
                    lines[i] = ''
                lines[end_lineno - 1] = ']' + tail
                lines_offset[end_lineno - 1] += len(
                    lines[end_lineno - 1]) - length_of_last_line

    injected_code = '\n'.join([
        f"__QULAB_TEMPLATE__ = \"{fname}\"",
        f"from qulab.executor.template import decode_mapping as __decode_{hash_str}",
        f"__VAR_{hash_str} = __decode_{hash_str}(\"{hash_str}\", \"\"\"",
        mapping_code, "    \"\"\")", *lines
    ])

    return injected_code, hash_str



import functools
import graphlib
import importlib
import os
from pathlib import Path

import click
from loguru import logger

from ..cli.config import get_config_value, log_options
from .load import (WorkflowType, find_unreferenced_workflows, get_entries,
                   load_workflow, make_graph)
from .schedule import CalibrationFailedError
from .schedule import maintain as maintain_workflow
from .schedule import run as run_workflow
from .transform import set_config_api
from .utils import workflow_template


@logger.catch(reraise=True)
def boot(script_path):
    """Run a script in a new terminal."""
    import subprocess
    import sys

    proc = subprocess.Popen([sys.executable, script_path])
    proc.communicate()


@logger.catch(reraise=True)
def check_toplogy(workflow: WorkflowType, code_path: str | Path) -> dict:
    graph = {}
    try:
        graphlib.TopologicalSorter(make_graph(workflow, graph,
                                              code_path)).static_order()
    except graphlib.CycleError as e:
        logger.error(
            f"Workflow {workflow.__workflow_id__} has a circular dependency: {e}"
        )
        raise e
    return graph


def command_option(command_name):
    """命令专属配置装饰器工厂"""

    def decorator(func):

        @click.option(
            '--code',
            '-c',
            default=lambda: get_config_value("code", str, command_name),
            help='The path of the code.')
        @click.option(
            '--data',
            '-d',
            default=lambda: get_config_value("data", str, command_name),
            help='The path of the data.')
        @click.option(
            '--api',
            '-a',
            default=lambda: get_config_value("api", str, command_name),
            help='The modlule name of the api.')
        @click.option(
            '--bootstrap',
            '-b',
            default=lambda: get_config_value("bootstrap", Path, command_name),
            help='The path of the bootstrap.')
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            bootstrap = kwargs.pop('bootstrap')
            if bootstrap is not None:
                boot(bootstrap)
            return func(*args, **kwargs)

        return wrapper

    return decorator


@click.command()
@click.argument('workflow')
@click.option('--code',
              '-c',
              default=lambda: get_config_value("code", str, 'create'),
              help='The path of the code.')
def create(workflow, code):
    """
    Create a new workflow file.
    """
    logger.info(f'[CMD]: create {workflow} --code {code}')
    if code is None:
        code = Path.cwd()

    fname = Path(code) / f'{workflow}'
    fname = Path(os.path.expanduser(fname))
    if fname.exists():
        click.echo(f'{workflow} already exists')
        return

    fname.parent.mkdir(parents=True, exist_ok=True)
    deps = find_unreferenced_workflows(code)

    with open(fname, 'w') as f:
        f.write(workflow_template(workflow, list(deps)))
    click.echo(f'{workflow} created')


@click.command()
@click.argument('key')
@click.argument('value', type=str)
@click.option('--api',
              '-a',
              default=lambda: get_config_value("api", str, 'set'),
              help='The modlule name of the api.')
def set(key, value, api):
    """
    Set a config.
    """
    logger.info(f'[CMD]: set {key} {value} --api {api}')
    from . import transform
    if api is not None:
        api = importlib.import_module(api)
        set_config_api(api.query_config, api.update_config, api.export_config)
    try:
        value = eval(value)
    except:
        pass
    transform.update_config({key: value})


@click.command()
@click.argument('key')
@click.option('--api',
              '-a',
              default=lambda: get_config_value("api", str, 'get'),
              help='The modlule name of the api.')
def get(key, api):
    """
    Get a config.
    """
    logger.info(f'[CMD]: get {key} --api {api}')
    from . import transform
    if api is not None:
        api = importlib.import_module(api)
        set_config_api(api.query_config, api.update_config, api.export_config)
    click.echo(transform.query_config(key))


@click.command()
@click.argument('workflow')
@click.option('--plot', '-p', is_flag=True, help='Plot the report.')
@click.option('--no-dependents',
              '-n',
              is_flag=True,
              help='Do not run dependents.')
@click.option('--retry', '-r', default=1, type=int, help='Retry times.')
@click.option('--freeze', is_flag=True, help='Freeze the config table.')
@log_options
@command_option('run')
def run(workflow, code, data, api, plot, no_dependents, retry, freeze):
    """
    Run a workflow.

    If the workflow has entries, run all entries.
    If `--no-dependents` is set, only run the workflow itself.
    If `--retry` is set, retry the workflow when calibration failed.
    If `--freeze` is set, freeze the config table.
    If `--plot` is set, plot the report.
    If `--api` is set, use the api to get and update the config table.
    If `--code` is not set, use the current working directory.
    If `--data` is not set, use the `logs` directory in the code path.
    """
    logger.info(
        f'[CMD]: run {workflow} --code {code} --data {data} --api {api}'
        f'{" --plot" if plot else ""}'
        f'{" --no-dependents" if no_dependents else ""}'
        f' --retry {retry}'
        f'{" --freeze " if freeze else ""}')
    if api is not None:
        api = importlib.import_module(api)
        set_config_api(api.query_config, api.update_config, api.export_config)
    if code is None:
        code = Path.cwd()
    if data is None:
        data = Path(code) / 'logs'

    code = Path(os.path.expanduser(code))
    data = Path(os.path.expanduser(data))

    wf = load_workflow(workflow, code)
    check_toplogy(wf, code)

    for i in range(retry):
        try:
            if no_dependents:
                if hasattr(wf, 'entries'):
                    for entry in get_entries(wf, code):
                        run_workflow(entry,
                                     code,
                                     data,
                                     plot=plot,
                                     freeze=freeze)
                else:
                    run_workflow(wf, code, data, plot=plot, freeze=freeze)
            else:
                if hasattr(wf, 'entries'):
                    for entry in get_entries(wf, code):
                        maintain_workflow(entry,
                                          code,
                                          data,
                                          run=True,
                                          plot=plot,
                                          freeze=freeze)
                else:
                    maintain_workflow(wf,
                                      code,
                                      data,
                                      run=True,
                                      plot=plot,
                                      freeze=freeze)
            break
        except CalibrationFailedError as e:
            if i == retry - 1:
                raise e
            logger.warning(f'Calibration failed, retrying ({i + 1}/{retry})')
            continue


@click.command()
@click.argument('workflow')
@click.option('--retry', '-r', default=1, type=int, help='Retry times.')
@click.option('--plot', '-p', is_flag=True, help='Plot the report.')
@log_options
@command_option('maintain')
def maintain(workflow, code, data, api, retry, plot):
    """
    Maintain a workflow.

    If the workflow has entries, run all entries.
    If `--retry` is set, retry the workflow when calibration failed.
    If `--plot` is set, plot the report.
    If `--api` is set, use the api to get and update the config table.
    If `--code` is not set, use the current working directory.
    If `--data` is not set, use the `logs` directory in the code path.
    """
    logger.info(
        f'[CMD]: maintain {workflow} --code {code} --data {data} --api {api}'
        f' --retry {retry}'
        f'{" --plot" if plot else ""}')
    if api is not None:
        api = importlib.import_module(api)
        set_config_api(api.query_config, api.update_config, api.export_config)
    if code is None:
        code = Path.cwd()
    if data is None:
        data = Path(code) / 'logs'

    code = Path(os.path.expanduser(code))
    data = Path(os.path.expanduser(data))

    wf = load_workflow(workflow, code)
    check_toplogy(wf, code)

    for i in range(retry):
        try:
            if hasattr(wf, 'entries'):
                for entry in get_entries(wf, code):
                    maintain_workflow(entry,
                                      code,
                                      data,
                                      run=False,
                                      plot=plot,
                                      freeze=False)
            else:
                maintain_workflow(wf,
                                  code,
                                  data,
                                  run=False,
                                  plot=plot,
                                  freeze=False)
            break
        except CalibrationFailedError as e:
            if i == retry - 1:
                raise e
            logger.warning(f'Calibration failed, retrying ({i + 1}/{retry})')
            continue
